#!/usr/bin/tcsh -f
set echo_style='bsd' ; setenv LC_ALL C #; set echo verbose

# Copyright 2012 Leon van Dommelen.  This software is made available
# under the GNU Public License version 3.  See folder copyrights.

# Purpose: Process a set of files to create new ones.
#          Maintains backups for the created files.
#          Trashes any file @tmp0.bat.

# Usage
goto start
use:
echo 'Usage: dofiles.bat LISTROOT FILES EXT2 SCRIPT PARS'
echo '       LISTROOT is the root file name for the list of files to convert.'
echo '          If LISTROOT starts with @, it must be quoted in DOS.'
echo '          LISTROOT.bat is created from FILES.'
echo '       FILES is the file specification of the files to convert.'
echo '          Applies to the current folder and its subfolders.'
echo '          Should not be quoted in DOS.'
echo '          If FILES is - a file list LISTROOT.lst must already exist'
echo '             In such a file list, a specification NAME.F0L is'
echo '             assumed to indicate a extensionless folder named NAME.'
echo '             Otherwise folders are not supported at this time.'
echo '       EXT2 is the extension of the created files.'
echo '          If EXT2 is * the source file extension is used.'
echo '          EXT2 should be * if the source file extension contains wildcards.'
echo '       SCRIPT is the single-file script to use to create a new file.'
echo '          May not be quoted in DOS.'
echo '          Normally SCRIPT will be called for each file NAME.EXT as'
echo '             SOURCE SCRIPT NAME EXT EXT2 PARS'
echo '          and is supposed to name the created file NAME@new.EXT2.  (That'
echo '          will get renamed NAME.EXT2 after backing up any existing file.)'
echo '          SCRIPT may receive additional nil parameters in unix.'
echo '          If SCRIPT is - dofiles exits just before the creation time;'
echo '          if SCRIPT is + dofiles resumes from just after that time.  That'
echo '          allows the NAME@new.EXT creation to be done outside dofiles.'
echo '          If SCRIPT is an actual script, it must be in system-files/bin.'
echo '          If SCRIPT returns a nonzero error status for any single file,'
echo '          all processing aborts and any created @new files are left as is.'
echo '       PARS are the parameters used by the script following NAME EXT EXT2.'
echo '          If dofiles is used in a runscript file, only three (four) such'
echo '          parameters can be accommodated at this time.  Otherwise there'
echo '          can be five.'
goto err
start:

# Backup model:
#
# Think of backups as layers.  The top layer is the current versions of
# the files.  Lower layers are the same set of files at earlier times.
# Managing backups means moving these layers, as a whole, up or down.
#
# A hole in a layer is simply a pointer that the file at that location
# is the same as the next more recent nonhole.  As an exception, a hole
# in a just-created @new layer means the file is the same as the one in
# the current layer.  That difference is accounted for when moving the
# layer of @new files into the current one: if there is no @new file,
# the original is not moved.  Orphaned @new files are not supposed to
# exist and are deleted on sight.
#
# To deal with variable set sizes consider that files not mentioned in
# the file list at that level do not exist.  If there is in fact a
# stray file, it will be deleted on sight.

# set L2HTOP (do not change the format of the next two lines)
setenv L2HTOP '/home/user/l2hdir'
if !(-f "$L2HTOP/install_l2h.desktop") goto errl2h

# check the basic files needed for execution
if !(-rx "$L2HTOP/bin/filechk.bat") goto errfils
"$L2HTOP/bin/filechk.bat"
if ($status) goto errfils
if !(-rx "$L2HTOP/system-files/bin/filechk.bat") goto errfils
"$L2HTOP/system-files/bin/filechk.bat"
if ($status) goto errfils
if !(-rx "$L2HTOP/system-files/data/filechk.bat") goto errfils
"$L2HTOP/system-files/data/filechk.bat"
if ($status) goto errfils

# if SCRIPT is + go do postprocessing
if ("$4" == '+') goto dopost

# check the validity of SCRIPT
if ("$4" == '-') goto scriptok
if !(-r "$L2HTOP/system-files/bin/$4") goto errscript
scriptok:

# initialize errlev
set errlev='0'

# see whether an existing list of files is to be used
if ("$2" != '-') goto mklst
if !(-f "$1.lst") goto errlst
goto chklst

# create the file list
mklst:
echo "$2" | fgrep '.'
set stat="$status"
if ($stat >= 2) goto errsys
if ($stat >= 1) goto errspec
if (-f "$1.lst") rm "$1.lst"
if (-e "$1.lst") goto errsys
find . -name "$2" -type f > "$1".lst
if ($status) goto errmklst
if !(-f "$1.lst") goto errsys

# check the list for file types
chklst:
cat "$1".lst | fgrep -v '.'
set stat="$status"
if ($stat >= 2) goto errsys
if ($stat >= 1) goto havpts
goto errpts
havpts:
cat "$1".lst | fgrep '.' > "$1".tmp
set stat="$status"
if ($stat >= 2) goto errsys
if ($stat >= 1) goto errnon
if (-f "$1.tmp") rm "$1.tmp"

# create the batch file
if (-f "$1.bat") rm "$1.bat"
if (-e "$1.bat") goto errsys
"$L2HTOP/bin/subbin.exe" "$L2HTOP/system-files/data/mkprocfils.sub" "$1".lst "$1".bat
if ($status) goto errmkbat
if !(-f "$1.bat") goto errmkbat

# get rid of orphaned @new files
set errlev='0'
echo
source "$1".bat "$L2HTOP/system-files/bin/file_extsel.bat" "$3" "$L2HTOP/system-files/bin/file_bcklvshw.bat" new
if ($status) goto errsys
if ("$errlev" == 0) goto enddelnw
echo
echo 'There are orphaned "@new" files, as listed above.  They are presumably'
echo 'left over from a failed (or undone) previous processing.  In any case,'
echo 'they will have to be removed to proceed.  (If you may want to keep them,'
echo 'move them out of the way to somewhere else *now*.)'
echo
"$L2HTOP/bin/choose.exe" qd "Press key d to delete the above files or q to quit processing"
set stat="$status"
if ($stat >= 3) goto errsys
if ($stat >= 2) goto dodelnw
goto errq
dodelnw:
source "$1".bat "$L2HTOP/system-files/bin/file_extsel.bat" "$3" "$L2HTOP/system-files/bin/file_bcklvrm.bat" new
if ($status) goto err
enddelnw:

# run the script with its parameters
if ("$4" == '-') goto done
"$L2HTOP/bin/clearold.bat" y @tmp0.bat
if ($status) goto err
set errlev='0'
source "$1".bat "$L2HTOP/system-files/bin/$4" "$3" "$5" "$6" "$7" "$8" "$9"
if ($status) goto errcnv
"$L2HTOP/bin/clearold.bat" y @tmp0.bat
if ($status) goto err

# postprocessing
dopost:
if !(-f "$1.bat") goto errgone

# see whether we need to do backups
source "$1".bat "$L2HTOP/system-files/bin/file_extsel.bat" "$3" "$L2HTOP/system-files/bin/file_bcklvchk0.bat"
if ($status) goto dobck
goto mvnew
dobck:

# see whether the @old level is already free
if !(-f "$1@old.bat") goto mv01
source "$1"@old.bat "$L2HTOP/system-files/bin/file_extsel.bat" "$3" "$L2HTOP/system-files/bin/file_bcklvchk.bat" old
if ($status) goto try2
rm "$1@old.bat"
if !(-f "$1@old.bat") goto mv01
echo "** Unable to remove expired backup file list $1@old.bat."
echo '   Please do that manually now or undo will be corrupted.'
"$L2HTOP/bin/choose.exe" anykey "Hit any key when this is done"
if !(-f "$1@old.bat") goto mv01
goto errrm1

# see whether the @older level is already free
try2:
if !(-f "$1@older.bat") goto mv12
source "$1"@older.bat "$L2HTOP/system-files/bin/file_extsel.bat" "$3" "$L2HTOP/system-files/bin/file_bcklvchk.bat" older
if ($status) goto try3
rm "$1@older.bat"
if !(-f "$1@older.bat") goto mv12
echo "** Unable to remove expired backup file list $1@older.bat."
echo '   Please do that manually now or undo will be corrupted.'
"$L2HTOP/bin/choose.exe" anykey "Hit any key when this is done"
if !(-f "$1@older.bat") goto mv12
goto errrm2

# see whether the @oldest level is already free
try3:
if !(-f "$1@oldest.bat") goto mv23
source "$1"@oldest.bat "$L2HTOP/system-files/bin/file_extsel.bat" "$3" "$L2HTOP/system-files/bin/file_bcklvchk.bat" oldest
if ($status) goto cln3
rm "$1@oldest.bat"
if !(-f "$1@oldest.bat") goto mv23
echo "** Unable to remove expired backup file list $1@oldest.bat."
echo '   Please do that manually now or undo will be corrupted.'
"$L2HTOP/bin/choose.exe" anykey "Hit any key when this is done"
if !(-f "$1@oldest.bat") goto mv23
goto errrm3

# remove the @oldest level
cln3:
source "$1"@oldest.bat "$L2HTOP/system-files/bin/file_extsel.bat" "$3" "$L2HTOP/system-files/bin/file_bcklvrm.bat" oldest
if ($status) goto errrm
rm "$1@oldest.bat"
if !(-f "$1@oldest.bat") goto mv23
echo "** Unable to remove expired backup file list $1@oldest.bat."
echo '   Please do that manually now or undo will be corrupted.'
"$L2HTOP/bin/choose.exe" anykey "Hit any key when this is done"
if !(-f "$1@oldest.bat") goto mv23
goto errrm3

# move the @older level into @oldest
mv23:
cp -p "$1@older.bat" "$1@oldest.bat"
if ($status) goto errsys
if !(-f "$1@oldest.bat") goto errsys
source "$1"@older.bat "$L2HTOP/system-files/bin/file_extsel.bat" "$3" "$L2HTOP/system-files/bin/file_bcklvrm.bat" oldest
if ($status) goto errstray
source "$1"@older.bat "$L2HTOP/system-files/bin/file_extsel.bat" "$3" "$L2HTOP/system-files/bin/file_bcklvmv.bat" older oldest
if ($status) goto errmv
rm "$1@older.bat"
if !(-f "$1@older.bat") goto mv12
echo "** Unable to remove expired backup file list $1@older.bat."
echo '   Please do that manually now or undo will be corrupted.'
"$L2HTOP/bin/choose.exe" anykey "Hit any key when this is done"
if !(-f "$1@older.bat") goto mv12
goto errrm2

# move the @old level into @older
mv12:
cp -p "$1@old.bat" "$1@older.bat"
if ($status) goto errsys
if !(-f "$1@older.bat") goto errsys
source "$1"@old.bat "$L2HTOP/system-files/bin/file_extsel.bat" "$3" "$L2HTOP/system-files/bin/file_bcklvrm.bat" older
if ($status) goto errstray
source "$1"@old.bat "$L2HTOP/system-files/bin/file_extsel.bat" "$3" "$L2HTOP/system-files/bin/file_bcklvmv.bat" old older
if ($status) goto errmv
rm "$1@old.bat"
if !(-f "$1@old.bat") goto mv01
echo "** Unable to remove expired backup file list $1@old.bat."
echo '   Please do that manually now or undo will be corrupted.'
"$L2HTOP/bin/choose.exe" anykey "Hit any key when this is done"
if !(-f "$1@old.bat") goto mv01
goto errrm1

# move the current files to be replaced into @old
mv01:
cp -p "$1.bat" "$1@old.bat"
if ($status) goto errsys
if !(-f "$1@old.bat") goto errsys
source "$1".bat "$L2HTOP/system-files/bin/file_extsel.bat" "$3" "$L2HTOP/system-files/bin/file_bcklvrm.bat" old
if ($status) goto errstray
source "$1".bat "$L2HTOP/system-files/bin/file_extsel.bat" "$3" "$L2HTOP/system-files/bin/file_bcklvmv0.bat" new old
if ($status) goto errmv

# move the @new files into the current level
mvnew:
source "$1".bat "$L2HTOP/system-files/bin/file_extsel.bat" "$3" "$L2HTOP/system-files/bin/file_bcklvmv1.bat" new
if ($status) goto errnew

# done
done:
goto end

# errors
errl2h:
echo "*** Fatal error: cannot find file $L2HTOP/install_l2h.desktop."
echo '    You may need to rerun install_l2h in the l2h folder.  Make sure'
echo '    that the current folder is somewhere inside the l2h folder while'
echo '    doing that.'
goto err
errfils:
echo '*** Fatal error: essential files are missing from the l2h folder'
echo '    or you do not have enough access to them.'
echo '    You may need to reinstall l2h from scratch.'
goto err
errscript:
echo "*** Program error: There is no $4 in $L2HTOP/system-files/bin or"
echo '    you do not have access to it.'
goto err
errlst:
echo "*** Program error: The file list $1.lst is missing"
goto err
errspec:
echo "*** Program error: file specification $2 contains no file type"
goto err
errmklst:
echo "*** Error searching for files matching $2"
goto err
errpts:
echo '*** Fatal error: There are files without filetypes in the list of files'
echo '      to convert.'
goto err
errnon:
echo '*** Fatal error: The list of files to convert is empty'
goto err
errmkbat:
echo '*** Unable to create a script to process the selected files.'
goto err
errcnv:
echo '*** Error processing the files.'
goto err
errgone:
echo "*** File list $1.bat is gone??"
goto err
errrm1:
echo "*** Backup file list $1@old.bat cannot be deleted.  Why?"
echo '    Check your permission to delete it.  Or a disk going bad?\!\!'
goto err
errrm2:
echo "*** Backup file list $1@older.bat cannot be deleted.  Why?"
echo '    Check your permission to delete it.  Or a disk going bad?\!\!'
goto err
errrm3:
echo "*** Backup file list $1@oldest.bat cannot be deleted.  Why?"
echo '    Check your permission to delete it.  Or a disk going bad?\!\!'
goto err
errrm:
echo '*** Unable to remove the oldest backups to make room for new ones.'
goto err
errstray:
echo '*** Unable to get rid of stray backups.'
goto err
errmv:
echo '*** Unable to move the older files to make room for the new ones.'
goto err
errnew:
echo '*** Unable to replace the old files by the newly created ones.'
goto err
errq:
echo '*** Aborted by user.'
goto err
errsys:
echo '*** Operating system error.  Check disk space and file protections.'
echo '    If not, try rerunning install_l2h in the l2h_folder.'
err:
exit 1

end:
