      subroutine hyphen(hyp,inmax,inchr,outdim,outmax,outchr)

c     Copyright 2012 Leon van Dommelen.  This software is made available
c     under the GNU Public License version 3.

c     Put discretionary hyphens in the string inchr and output the
c     result as string outchr.  NonASCII letters in the string inchr
c     must be encoded in internal format.  Multiple \V characters at the
c     end of an internal string, (for distinguishing different types),
c     are acceptable.

c     This subroutine requires that a suitable hyphenation pattern file
c     is loaded using a call with inmax negative.  This file must hold
c     the letters of the alphabet, with their lowercase equivalents.  It
c     must also hold the hyphenation patterns needed for the TeX
c     hyphenation algorithm.  And it may hold a list of exception words.
c     See the examples for the correct format.

c     If there is a problem with the hyphenation, normally inchr is
c     simply copied to outchr without changes.  Note that words longer
c     than 127 bytes, when written in UTF-8 format, will not be
c     hyphenated.  Sorry about that.  To see why specific words are not
c     hyphenated requires that the -V option is used on the subbin.exe
c     command.

c     Some notes in case you want to modify this subroutine:
c
c     This subroutine encodes character strings into a database datisu
c     of 4 byte integers for speed.  On a 8 byte machine configuration,
c     you may be able to increase speed further by switching to 8 byte
c     integers.  To do so, in *all* subroutines in this file change
c     integer*4 into integer*8 and cpi=4 to cpi=8.  However, the change
c     will result in an increase in storage.
c
c     One potential source of confusion is the term "character."  What
c     Fortran calls a character is really a byte, not a character.  A
c     single Unicode character is often several bytes long.  And that
c     depends on the encoding scheme too.  Don't you miss those good old
c     days of just US-ASCII?

      implicit none

c input:

c     the discretionary hyphen to insert, or the data file name when inmax < 0
      character*(*) hyp

c     number of characters in the word to hyphenate; if this is negative, a
c     the subroutine will merely read in hyphenation data file hyp and exit
      integer inmax
c     the word to hyphenate;
c     special characters must be in \U\l...\V internal format, except
c     that a \V immediately following an \V is ignored.
      character*1 inchr(*)

c     available size of array out
      integer outdim

c output:

c     number of characters in the hyphenated word
      integer outmax

c     hypenated word
      character*1 outchr(outdim)

c common

c     verbosity
      integer verbos
c     substitions file
      character*132 subfil
      integer subfll
c     line count of the substitutions file
      integer lincts
c     text file
      character*132 txtfil
      integer txtfll
c     line count of the text file
      integer linctt

c     debug block
      common/cmdbg/verbos,subfil,subfll,lincts,txtfil,txtfll,linctt

c external

      integer utf8lt
      external utf8lt

c local data storage:

c     declared number of letters in the language
      integer letdim
      parameter (letdim=350)
c     actual number of letters
      integer letmax
c     letter index and its search interval
      integer let,letlo,lethi
c     letter character numbers
      integer letcno(letdim)
c     declared number of corresponding lowercase letters
      integer lcdim
      parameter (lcdim=letdim)
c     the lower case character list
      integer lccno(lcdim)
c     list index and count
      integer lc,lcct
c     identity mappings
      integer letid
c     end position of the lowercase character(s) in terms of let
      integer letlcm(0:letdim)
      data letlcm(0)/0/
      save letmax,letcno,lccno,letlcm

c     declared number of integers in the database (hu: 100000? no: 35000?)
      integer datdim
      parameter (datdim=50000)
c     actual number of integers in the database
      integer datmax
c     data index
      integer dat,dat2
c     the database as integer storage units, and a couple of selected values
      integer*4 datisu(datdim),datiu,datiu2
c     characters per integer in datisu as declared above
      integer cpi
      parameter (cpi=4)
c     hyphenation values per integer
      integer hvlpi
      parameter (hvlpi=cpi*2-1)
c     hyphens per integer
      integer hyppi
      parameter (hyppi=cpi*8-1)
c     sublocation in a data value
      integer sub
      save datmax,datisu

c     declared number of patterns (hu: 65000? no: 30000?)
      integer patdim
      parameter (patdim=15000)
c     actual number of patterns
      integer patmax
c     index of the sorted patterns
      integer pat
c     index pun of the unsorted patterns in terms of pat
      integer patpun(patdim)
c     pattern location in datisu in terms of pun
      integer pundat(patdim)
c     beginning of patterns of a given length
      integer pltpat(128),pltmin,pltmax,plt
      save patmax,patpun,pundat,pltpat,pltmin,pltmax

c     declared number of exception words
      integer ewddim
      parameter (ewddim=10000)
c     actual number of exception words
      integer ewdmax
c     index of the sorted words
      integer ewd
c     index eun of the unsorted words in terms of ewd
      integer ewdeun(ewddim)
c     word location in datisu in terms of eun
      integer eundat(ewddim)
c     beginning of words of a given length
      integer eltewd(128),eltmin,eltmax,elt
      save ewdmax,ewdeun,eundat,eltewd,eltmin,eltmax

c     first and last hyphenation point
      integer hypbeg,hypend
      save hypbeg,hypend

c local other:

c     maximum unicode character number
      integer uclim
      parameter (uclim=1114111)

c     whether we have initialized
      integer init
      data init/0/
      save init

c     declared number of unicode character numbers in the input string
      integer ucdim
      parameter (ucdim=127)
c     actual number of unicode character numbers in the input string
      integer ucmax
c     unicode character number index
      integer uc
c     unicode character numbers
      integer uccno(ucdim)
c     character in inchr that this character ends
      integer ucinmx(0:ucdim)

c     declared number of characters in the UTF-8 string to look for
      integer fnddim
      parameter (fnddim=127)
c     actual number of characters in the string to look for
      integer fndmax
c     character index
      integer fnd
c     characters in the string
      character*1 fndchr(0:fnddim+1)
c     unicode character number index that the character corresponds to
      integer fnduc(fnddim)
c     the hyphenation values
      integer fndhvl(-1:fnddim+1)

c     declared number of characters in the work string
      integer strdim
      parameter (strdim=255)
c     actual number of characters in the work string
      integer strmax
c     character indices and their starting value
      integer str,str2,str0
c     characters in the string
      character*1 strchr(strdim)
c     the string as a single character string
      character*(strdim) string
      equivalence (string,strchr(1))
c     hyphenation values between the characters
      integer strhvl(0:127)

c     ascii values of 0, 9, and -
      integer asc_0,asc_9,asc_mn
      parameter (asc_0=48,asc_9=57,asc_mn=45)
c     ascii values of space, carriage return, and linefeed
      integer asc_tb,asc_sp,asc_cr,asc_lf
      parameter (asc_tb=9,asc_sp=32,asc_cr=13,asc_lf=10)
c     ascii values of a, z, A, and Z
      integer asc_a,asc_z
      parameter (asc_a=97,asc_z=122)
c     ascii value of the internal word start and end
      integer asc_i1,asc_i2
      parameter (asc_i1=21,asc_i2=22)

c     line count in the hyphenation data file
      integer linct
c     read mode
      integer rdmode
c     its values (casefolds, patterns, exception words)
      integer rdflds,rdpats,rdewds
      parameter (rdflds=1,rdpats=2,rdewds=3)
c     whether the line has TeX commands or brackets
      integer hastex
c     whether we are in a comment
      integer incom

c     whether the previous character was a digit
      integer prvdig

c     whether the word has non-letters
      integer nonlet

c     counts of good and bad hyphenation points
      integer ghypct,bhypct

c     packed length of the string
      integer pcklt

c     input string character indices
      integer in,in0

c     input string character index
      integer out

c     unicode character number
      integer cno

c     integer as string
      character*12 cint,cint2
      integer cintl,cint2l

c     encoded character length
      integer enclt

c     error message
      character*76 msg

c     character
      character*1 chr
      integer ichr

c     generic index
      integer i

c executable

c     formats
 10   format(a,a,a,a,a)
 20   format(a,2048a1)

c     note where we are
      call insub('hyphen')

c     initialize
      if(inmax.lt.0)then
         init=1

c        open the data file
         call openru(hyp,7)

c        set defaults
         hypbeg=0
         hypend=0

c        initialize the reading
         rdmode=0
         letmax=0
         letid=0
         patmax=0
         ewdmax=0
         dat=0

c        next line
         linct=0
 1000    linct=linct+1
         hastex=0
         incom=0
         strmax=0

c        next character in line
 1010    call getchr(7,chr,*1900,*9100)
         ichr=ichar(chr)

c        handle end of line
         if(ichr.eq.asc_lf)then

c           ignore blank lines and trailing spaces
            if(strmax.eq.0)goto 1000
            if(strchr(strmax).eq.' ')strmax=strmax-1
            if(strmax.eq.0)goto 1000

c           check for tex commands
            if(hastex.ne.0)then
c              check for tex commands that change read mode
               if(index(string(1:strmax),'\\casefolds').ne.0)then
                  if(rdmode.ne.rdflds-1 .and. rdmode.ne.rdflds)goto 9101
                  rdmode=rdflds
               else if(index(string(1:strmax),'\\patterns').ne.0)then
                  if(rdmode.ne.rdpats-1 .and. rdmode.ne.rdpats)goto 9102
                  rdmode=rdpats
               else if(index(string(1:strmax),'\\hyphenation').ne.0)then
                  if(rdmode.ne.rdewds-1 .and. rdmode.ne.rdewds)goto 9103
                  rdmode=rdewds
               endif
               i=index(string(1:strmax),'\\firsthyphen')
               if(i.ne.0)then
                  i=i+12
                  if(strchr(i).ne.'{')goto 9104
                  if(strmax-i.lt.2)goto 9104
                  enclt=index(string(i+1:strmax),'}')-1
                  if(enclt.le.0)goto 9104
                  call rbaseb(10,asc_0,enclt,strchr(i+1),hypbeg)
                  if(hypbeg.le.0)goto 9104
               endif
               i=index(string(1:strmax),'\\lasthyphen')
               if(i.ne.0)then
                  i=i+11
                  if(strchr(i).ne.'{')goto 9105
                  if(strmax-i.lt.2)goto 9105
                  enclt=index(string(i+1:strmax),'}')-1
                  if(enclt.le.0)goto 9105
                  call rbaseb(10,asc_0,enclt,strchr(i+1),hypend)
                  if(hypend.le.0)goto 9105
               endif
c              in any case not a data line
               goto 1000
            endif

c           add to the appropriate database
            if(rdmode.eq.rdflds)then

c              casefolds line

c              next casefold
               if(letmax.ge.letdim)goto 9119
               letmax=letmax+1
               lc=letlcm(letmax-1)
               lcct=-1

c              see whether there is a mapping
               str0=
     &          index(string(1:strmax),char(226)//char(134)//char(146))
               if(str0.gt.0)then
                  enclt=0
                  call rutf8(str0-1,strchr,enclt,cno)
                  if(cno.le.32)goto 9115
                  if(enclt.ne.str0-1)goto 9110
                  letcno(letmax)=cno
                  str0=str0+2
                  if(str0.ge.strmax)goto 9111
                  if(strchr(str0+1).eq.' ')goto 9112
                  lcct=0
 1110             enclt=0
                  call rutf8(strmax-str0,strchr(str0+1),enclt,cno)
                  if(cno.le.32)goto 9115
                  lcct=lcct+1
                  if(lc+lcct.gt.lcdim)goto 9118
                  lccno(lc+lcct)=cno
                  str0=str0+enclt
                  if(str0.ge.strmax)goto 9111
                  if(strchr(str0+1).ne.' ')goto 1110
                  str0=str0+1
                  if(str0.ge.strmax)goto 9111
               endif
               if(index(string(str0+1:strmax),' ').ne.0)goto 9114
               enclt=index(string(str0+1:strmax),',')-1
               if(enclt.le.0)goto 9110
               call rbaseb(10,asc_0,enclt,strchr(str0+1),cno)
               if(cno.le.32)goto 9116
               if(lcct.ge.0)then
                  if(letcno(letmax).ne.cno)goto 9113
               else
                  letcno(letmax)=cno
               endif
               if(letmax.gt.1)then
                  if(letcno(letmax).le.letcno(letmax-1))goto 9109
               endif
               str0=str0+enclt+1
               if(str0.ge.strmax)goto 9111
 1150          enclt=index(string(str0+1:strmax),',')-1
               if(enclt.le.0)enclt=strmax-str0
               call rbaseb(10,asc_0,enclt,strchr(str0+1),cno)
               if(cno.le.32)goto 9116
               lc=lc+1
               if(lcct.gt.0)then
                  lcct=lcct-1
                  if(lccno(lc).ne.cno)goto 9113
               else if(lcct.eq.0)then
                  goto 9113
               else
                  if(lc.gt.lcdim)goto 9118
                  lccno(lc)=cno
               endif
               str0=str0+enclt+1
               if(str0.lt.strmax)goto 1150
               if(str0.ne.strmax+1)goto 9110
               if(lcct.gt.0)goto 9113
               letlcm(letmax)=lc
               if(letlcm(letmax).eq.letlcm(letmax-1)+1)then
                  if(letcno(letmax).eq.lccno(letlcm(letmax)))
     &             letid=letid+1
               endif
               goto 1000

            else if(rdmode.eq.rdpats)then

c              patterns line

c              next pattern
               if(patmax.ge.patdim)goto 9129
               patmax=patmax+1
               patpun(patmax)=patmax
               pundat(patmax)=dat+1

c              take apart in characters and hyphenation values
               str2=0
               prvdig=0
               strhvl(str2)=0
               do 1210 str=1,strmax
                  ichr=ichar(strchr(str))
                  if(ichr.eq.asc_sp)goto 9121
                  if(ichr.eq.ichar('.')
     &             .and. str.ne.1 .and. str.ne.strmax)goto 9122
                  if(ichr.lt.asc_sp .or. ichr.eq.127)goto 9123
                  if(ichr.ge.asc_0 .and. ichr.le.asc_9
     &             .and. prvdig.eq.0)then
                     strhvl(str2)=ichr-asc_0
                     prvdig=1
                  else
                     if(str2.ge.127)goto 9125
                     str2=str2+1
                     strchr(str2)=char(ichr)
                     strhvl(str2)=0
                     prvdig=0
                  endif
 1210          continue
               strmax=str2
               if(strmax.le.0)goto 9126

c              store the characters
               dat2=dat
               call pckcs(strmax,strchr,1,datdim,dat2,datisu)
               if(dat2.lt.0)goto 9140
               dat=dat2

c              store the hyphenation values
               if(dat+(strmax+1+hvlpi-1)/hvlpi.gt.datdim)goto 9140
               sub=hvlpi
               do 1220 str=0,strmax
                  sub=sub+1
                  if(sub.gt.hvlpi)then
                     dat=dat+1
                     datisu(dat)=0
                     sub=1
                     datiu=1
                  else
                     datiu=datiu*16
                  endif
                  datisu(dat)=datisu(dat)+strhvl(str)*datiu
 1220          continue

            else if(rdmode.eq.rdewds)then

c              exception words line

c              next exception words
               if(ewdmax.ge.ewddim)goto 9139
               ewdmax=ewdmax+1
               ewdeun(ewdmax)=ewdmax
               eundat(ewdmax)=dat+1

c              take apart in characters and hyphens
               str2=0
               strhvl(str2)=0
               do 1310 str=1,strmax
                  ichr=ichar(strchr(str))
                  if(ichr.eq.asc_mn)then
                     if(str.eq.1 .or. str.eq.strmax)goto 9131
                     if(strhvl(str2).ne.0)goto 9132
                     strhvl(str2)=1
                  else
                     if(str2.ge.127)goto 9135
                     str2=str2+1
                     strchr(str2)=char(ichr)
                     strhvl(str2)=0
                  endif
 1310          continue
               strmax=str2
               if(strmax.le.1)goto 9136

c              store the characters
               dat2=dat
               call pckcs(strmax,strchr,1,datdim,dat2,datisu)
               if(dat2.lt.0)goto 9140
               dat=dat2

c              store the hyphenation points
               if(dat+(strmax-1+hyppi-1)/hyppi.gt.datdim)goto 9140
               sub=hyppi
               do 1320 str=1,strmax-1
                  sub=sub+1
                  if(sub.gt.hyppi)then
                     dat=dat+1
                     datisu(dat)=0
                     sub=1
                     datiu=1
                  else
                     datiu=datiu*2
                  endif
                  datisu(dat)=datisu(dat)+strhvl(str)*datiu
 1320          continue

            else

            endif

c           start the next line
            goto 1000
         endif

c        inside a comment ignore everything except end-of-line
         if(incom.ne.0)goto 1010
c        look for start of comments
         if(ichr.eq.ichar('%'))then
            incom=1
            goto 1010
         endif

c        ignore leading and multiple whitespace
         if(ichr.eq.asc_cr)goto 1010
         if(ichr.eq.asc_sp .or. ichr.eq.asc_tb)then
            if(strmax.eq.0)goto 1010
            if(strchr(strmax).eq.' ')goto 1010
            ichr=asc_sp
         endif

c        check for special characters
         if(ichr.eq.ichar('{') .or. ichr.eq.ichar('}')
     &     .or. ichr.eq.ichar('\\'))hastex=1

c        add the character to the read string
         strmax=strmax+1
         if(strmax.gt.strdim)goto 9108
         strchr(strmax)=char(ichr)
         goto 1010

c        close the input file
 1900    call closu(7,hyp)
         datmax=dat

c        check for loose ends
         if(hypbeg.le.0)goto 9106
         if(hypend.le.0)goto 9107
         if(letmax.le.0)goto 9117
         if(patmax.le.0)goto 9127

c        log it
         if(verbos.gt.0)then
            print10
            print10,'Read in hyphen.txt:'
            call inttos(letmax,cint,cintl)
            call inttos(letid,cint2,cint2l)
            print10,'   ',cint(1:cintl),' letters, of which ',
     &       cint2(1:cint2l),' are normal lowercase ones'
            call inttos(patmax,cint,cintl)
            i=datmax
            if(ewdmax.gt.0)i=eundat(1)-1
            call inttos(i,cint2,cint2l)
            print10,'   ',cint(1:cintl),' patterns, taking ',
     &       cint2(1:cint2l),' storage locations'
            call inttos(ewdmax,cint,cintl)
            i=0
            if(ewdmax.gt.0)i=datmax-eundat(1)+1
            call inttos(i,cint2,cint2l)
            print10,'   ',cint(1:cintl),' exception words, taking ',
     &       cint2(1:cint2l),' storage locations'
            call inttos(datmax,cint,cintl)
            call inttos(datdim,cint2,cint2l)
            print10,'   total ',cint(1:cintl),
     &       ' storage locations out of a maximum of ',
     &       cint2(1:cint2l)
         endif

c        sort the patterns
         call srtdat(0,datdim,datisu,patdim,patmax,pundat,
     &    patpun,pltpat,pltmin,pltmax)

c        display histogram
         if(verbos.ge.2)print1930,(pltpat(plt+1)-pltpat(plt),plt=1,127)
 1930    format('Pattern length histogram:'/('   ',15i5))

c        sort the exception words
         call srtdat(0,datdim,datisu,ewddim,ewdmax,eundat,
     &    ewdeun,eltewd,eltmin,eltmax)

c        display histogram
         if(verbos.ge.2)print1940,(eltewd(elt+1)-eltewd(elt),elt=1,127)
 1940    format('Exception word length histogram:'/('   ',15i5))

c        done with initialization
         goto 9900
      endif

c     make sure that we are initialized
      if(init.eq.0)
     & call fatal('Program error: no hyphenation data read in')



c     read the input word

c     done if less than two characters
      if(inmax.le.1)goto 9800

c     initialize decoding the input word
      nonlet=0
      ucmax=0
      fndmax=0
      in=0
      ucinmx(0)=0

c     next input word character
 2000 in=in+1
      if(in.gt.inmax)goto 2900
      ichr=ichar(inchr(in))
c     take care of encoded characters
      if(ichr.eq.asc_i1)then
c        internal encoded character
         in0=in+1
 2200    in=in+1
         if(in.gt.inmax)goto 9220
         ichr=ichar(inchr(in))
         if(ichr.ne.asc_i2)goto 2200
         call rbaseb(26,asc_a,in-in0,inchr(in0),cno)
         if(cno.lt.0 .or. cno.gt.uclim)goto 9220
 2250    if(in.ge.inmax)goto 2500
         if(ichar(inchr(in+1)).ne.asc_i2)goto 2500
         in=in+1
         goto 2250
      endif
c     nonencoded character
      if(ichr.eq.32)goto 9230
      if(ichr.lt.32 .or. ichr.eq.127)goto 9240
      if(ichr.ge.128)goto 9250
      cno=ichr
c     soft hyphens are out
 2500 if(cno.eq.173)goto 9260
c     add the character to the unicode list
      if(ucmax.ge.ucdim)goto 9210
      ucmax=ucmax+1
      uccno(ucmax)=cno
      ucinmx(ucmax)=in
c     find the character in the letter list
      if(cno.lt.letcno(1))goto 2850
      if(cno.gt.letcno(letmax))goto 2850
      let=1
      if(cno.eq.letcno(let))goto 2700
      letlo=let
      let=letmax
      if(cno.eq.letcno(let))goto 2700
      lethi=let
 2600 let=letlo+(lethi-letlo)/2
      if(let.eq.letlo)goto 2850
      if(letcno(let).lt.cno)then
         letlo=let
         goto 2600
      else if(letcno(let).gt.cno)then
         lethi=let
         goto 2600
      else
         goto 2700
      endif
c     add its lowercase conversion to the string to search for
 2700 do 2800 lc=letlcm(let-1)+1,letlcm(let)
         cno=lccno(lc)
         if(fndmax+utf8lt(cno).gt.fnddim)goto 9210
         call wutf8(cno,fnddim-fndmax,enclt,fndchr(fndmax+1))
         do 2750 fnd=fndmax+1,fndmax+enclt
            fnduc(fnd)=ucmax
 2750    continue
         fndmax=fndmax+enclt
 2800 continue
      goto 2000
c     this is not a letter, but it might be in an exception word
 2850 nonlet=nonlet+1
      if(fndmax.ge.fnddim)goto 9210
      fndmax=fndmax+1
      fndchr(fndmax)=char(cno)
      fnduc(fndmax)=ucmax
      goto 2000
c     done reading the input word
 2900 continue

c     try the exception word list

c     search the exception word list
      call fnddat(datdim,datisu,ewddim,ewdmax,eundat,ewdeun,
     & eltewd,eltmin,eltmax,fnddim,fndmax,fndchr(1),pcklt,ewd)

c     if not found, try the patterns
      if(ewd.le.0)goto 4000

c     location of the hyphenation points in datisu
      dat=eundat(ewdeun(ewd))+pcklt

c     restore the hyphenation points
      dat=dat-1
      sub=hyppi
      do 3100 fnd=1,fndmax-1
         sub=sub+1
         if(sub.gt.hyppi)then
            dat=dat+1
            sub=1
            datiu=datisu(dat)
         endif
         datiu2=datiu/2
         fndhvl(fnd)=datiu-datiu2*2
         datiu=datiu2
 3100 continue

c     go output the string with hyphenation points
      goto 5000

c     try the hyphenation algorithm

c     no hyphenation possible if too short
 4000 if(ucmax.lt.hypbeg+hypend)goto 9800

c     if there are nonletters, the patterns are invalid
      if(nonlet.ne.0)goto 9270

c     initialize
      fndchr(0)='.'
      fndchr(fndmax+1)='.'
      do 4100 fnd=-1,fndmax+1
         fndhvl(fnd)=0
 4100 continue

c     find all patterns
      do 4500 plt=pltmin,pltmax
         do 4400 fnd=0,fndmax+2-plt
c           find the pattern
            call fnddat(datdim,datisu,patdim,patmax,pundat,patpun,
     &       pltpat,pltmin,pltmax,plt,plt,fndchr(fnd),pcklt,pat)
            if(pat.le.0)goto 4400
c           location of the hyphenation values in datisu
            dat=pundat(patpun(pat))+pcklt
c           unpack the values
            dat=dat-1
            sub=hvlpi
            do 4300 str=0,plt
               sub=sub+1
               if(sub.gt.hvlpi)then
                  dat=dat+1
                  sub=1
                  datiu=datisu(dat)
               endif
               datiu2=datiu/16
               i=datiu-datiu2*16
               datiu=datiu2
               if(i.eq.0)goto 4300
               if(i.gt.fndhvl(fnd+str-1))fndhvl(fnd+str-1)=i
 4300       continue
 4400    continue
 4500 continue

c     enforce hyphenation boundaries
      do 4510 fnd=1,fndmax-1
         if(fnduc(fnd+1).le.hypbeg)fndhvl(fnd)=0
         if(fnduc(fnd).gt.ucmax-hypend)fndhvl(fnd)=0
 4510 continue

c     go output the string with hyphenation points
      goto 5000

c     output the input string with hyphenation points

c     initialize the input and output indices
 5000 out=0
      in0=0

c     output the string with hyphenation points
      ghypct=0
      bhypct=0
      do 5500 fnd=1,fndmax-1
         if(fndhvl(fnd).gt.1)fndhvl(fnd)=mod(fndhvl(fnd),2)
         if(fndhvl(fnd).eq.0)goto 5500
         uc=fnduc(fnd)
         if(uc.eq.fnduc(fnd+1))then
            bhypct=bhypct+1
            goto 5500
         endif
         ghypct=ghypct+1
         do 5200 in=in0+1,ucinmx(uc)
            out=out+1
            if(out.gt.outdim)goto 9510
            outchr(out)=inchr(in)
 5200    continue
         in0=ucinmx(uc)
         do 5300 i=1,len(hyp)
            out=out+1
            if(out.gt.outdim)goto 9510
            outchr(out)=hyp(i:i)
 5300    continue
 5500 continue
      do 5600 in=in0+1,inmax
         out=out+1
         if(out.gt.outdim)goto 9510
         outchr(out)=inchr(in)
 5600 continue
      outmax=out

c     all done
      if(bhypct.gt.0 .or. ghypct.le.0)goto 9550
      goto 9900

c     various errors with the hyphenation data file
 9100 print10,'While reading hyphenation data file',hyp
      call fatal('Disk error reading the hyphenation data file')
 9101 msg='\\casefolds command out of order'
      goto 9150
 9102 msg='\\patterns command out of order'
      goto 9150
 9103 msg='\\hyphenation command out of order'
      goto 9150
 9104 msg='incorrect \\firsthyphen command'
      goto 9150
 9105 msg='incorrect \\lasthyphen command'
      goto 9150
 9106 msg='missing \\firsthyphen command'
      goto 9150
 9107 msg='missing \\lasthyphen command'
      goto 9150
 9108 msg='line is too long'
      goto 9150
c     casefold line problems
 9109 msg='character numbers in casefolds must be in increasing order'
      goto 9150
 9110 msg='invalid casefolds line format'
      goto 9150
 9111 msg='incomplete casefolds line'
      goto 9150
 9112 msg='incomplete UTF-8 character mapping'
      goto 9150
 9113 msg='UTF-8 character mapping and unicode numbers do not match'
      goto 9150
 9114 msg='the unicode numbers list may not contain spaces'
      goto 9150
 9115 msg='invalid UTF-8 character'
      goto 9150
 9116 msg='invalid unicode character number'
      goto 9150
 9117 msg='the data file contains no casefolds'
      goto 9150
 9118 print10,'Too many casefold lowercase letters: lcdim exceeded.'
      goto 9144
 9119 print10,'Too many casefolds: letdim exceeded.'
      goto 9144
c     pattern line problems
 9121 msg='patterns may not contain spaces'
      goto 9150
 9122 msg='points may only appear at the start or end of the pattern'
      goto 9150
 9123 msg='patterns may not contain control characters'
      goto 9150
 9125 msg='patterns longer than 127 characters are not supported'
      goto 9150
 9126 msg='a pattern has no characters'
      goto 9150
 9127 msg='the data file contains no patterns'
      goto 9150
 9129 print10,'Too many patterns: patdim exceeded.'
      goto 9144
c     exception word line problems
 9131 msg='hyphens may not appear at the start or end of the word'
      goto 9150
 9132 msg='duplicate hyphen in the exception word'
      goto 9150
 9135 msg='exception words longer than 127 characters are not supported'
      goto 9150
 9136 msg='exception words must have at least two characters'
      goto 9150
 9139 print10,'Too many exception words, ewddim exceeded'
      goto 9144
c     data storage ran out
 9140 print10
      print10,'While reading hyphenation data file',hyp
      print10,
     & 'Out of hyphenation data storage: datdim exceeded.  So far:'
      call inttos(patmax,cint,cintl)
      i=dat
      if(ewdmax.gt.0)i=eundat(1)-1
      call inttos(i,cint2,cint2l)
      print10,'   ',cint(1:cintl),' patterns, taking ',
     & cint2(1:cint2l),' storage locations'
      call inttos(ewdmax,cint,cintl)
      i=0
      if(ewdmax.gt.0)i=dat-eundat(1)+1
      call inttos(i,cint2,cint2l)
      print10,'   ',cint(1:cintl),' exception words, taking ',
     & cint2(1:cint2l),' storage locations'
      call inttos(dat,cint,cintl)
      call inttos(datdim,cint2,cint2l)
      print10,'   total ',cint(1:cintl),
     & ' storage locations out of a maximum of ',cint2(1:cint2l)
 9144 call fatal('Out of storage as described above.')
c     process the message
 9150 print10
      call inttos(linct,cint,cintl)
      print10,'While reading hyphenation data file',hyp,
     & ' line ',cint(1:cintl),':'
      do 9160 str=1,strmax
         ichr=ichar(strchr(str))
         if(ichr.lt.32 .or. ichr.eq.127)strchr(str)='?'
 9160 continue
      if(strmax.gt.0)print20,'"',(strchr(str),str=1,strmax),'"'
      print10,'Problem description:'
      do 9170 i=76,1,-1
         if(msg(i:i).gt.' ')goto 9171
 9170 continue
      print10,'   none?!'
      goto 9172
 9171 print10,'   ',msg(1:i)
 9172 subfll=0
      txtfll=0
      call fatal('Fix the above problem and try again')

c     input word problems
 9210 if(verbos.lt.2)goto 9800
      msg='the word to hyphenate is too long'
      goto 9280
 9220 if(verbos.lt.2)goto 9800
      msg='the word has a bad internal character encoding'
      goto 9280
 9230 if(verbos.lt.2)goto 9800
      msg='the word contains spaces'
      goto 9280
 9240 if(verbos.lt.2)goto 9800
      msg='the word contains control characters'
      goto 9280
 9250 if(verbos.lt.2)goto 9800
      msg='the word contains 8 bit characters'
      goto 9280
 9260 if(verbos.lt.2)goto 9800
      msg='the word contains soft hyphens'
      goto 9280
 9270 if(verbos.lt.2)goto 9800
      msg='the word has nonletters'
      goto 9280
 9280 call encstr(inmax,inchr,56,strmax,strchr)
      print10
      if(strmax.le.55)print20,'Word not hyphenated: "',
     & (strchr(str),str=1,strmax),'"'
      if(strmax.gt.55)print20,'Word not hyphenated: "...',
     & (strchr(str),str=strmax-51,strmax),'"'
      if(fndmax.gt.0)then
         do 9285 fnd=1,fndmax
            ichr=ichar(fndchr(fnd))
            if(ichr.lt.32 .or. ichr.eq.127)fndchr(fnd)='?'
 9285    continue
         print20,'Decoded: ',(fndchr(fnd),fnd=1,fndmax),'.','.','.'
      endif
      do 9290 i=76,1,-1
         if(msg(i:i).gt.' ')goto 9291
 9290 continue
      print10,'Problem: none?!'
      goto 9800
 9291 print10,'Problem: ',msg(1:i)
      goto 9800

c     problems while inserting hyphens
 9510 call fatal('Out of storage while hyphenating')
 9550 if(verbos.lt.2)goto 9900
      do 9555 fnd=1,fndmax
         ichr=ichar(fndchr(fnd))
         if(ichr.lt.32 .or. ichr.eq.127)fndchr(fnd)='?'
 9555 continue
      print10
      print20,'Hyphenating: ',(fndchr(fnd),fnd=1,fndmax)
      if(bhypct.gt.0)
     & print20,'Bad hyphenation points (inside a character) found'
      if(ghypct.le.0)
     & print20,'No acceptable hyphenation points found'
      goto 9900

c     return the input string unchanged
 9800 if(inmax.gt.outdim)call fatal('Out of storage')
      do 9850 str=1,inmax
         outchr(str)=inchr(str)
 9850 continue
      outmax=inmax

c     return
 9900 call outsub('hyphen')
      return
      end



      subroutine pckcs(strmax,strchr,addsiz,datdim,dat,datisu)

c     Pack the string strchr of strmax characters into integers in the
c     database datisu.  Returns dat=-1 if out of storage.

      implicit none

c input

c     length of the string
      integer strmax
c     the string characters
      character*1 strchr(*)

c     if nonzero, put strmax in the leading position of the packed string
      integer addsiz

c     declared number of integers in the database
      integer datdim

c input/output

c     last filled position in datisu
      integer dat

c     the database as integer storage units, and a couple of selected values
      integer*4 datisu(datdim),datint,datiu
c     characters per integer in datisu as declared above
      integer cpi
      parameter (cpi=4)
c     datint divided into individual characters
      character*1 datchr(cpi)
      equivalence (datchr(1),datint)

c local

c     endianness
      integer endian
      data endian/-1/
      save endian

c     position in strchr
      integer str

c     character position in a packed integer
      integer ic

c executable

c     note where we are
      call insub('pckcs')

c     nothing to do when no characters to pack.
      if(addsiz.eq.0 .and. strmax.eq.0)goto 9900

c     negative number of characters is not possible
      if(strmax.lt.0)call fatal(
     & 'Program error: negative number of characters in pckcs')

c     cannot mark more than 127 characters
      if(addsiz.ne.0 .and. strmax.gt.127)call fatal(
     & 'Program error: stored-size string more than 127 characters')

c     check for endianness
      if(endian.eq.-1)then
         datiu=256
         datiu=127*datiu**(cpi-1)
         ic=1
         endian=0
         datint=0
         datchr(endian*(cpi+1)+(1-2*endian)*ic)=char(127)
         if(datint.ne.datiu)then
            endian=1
            datint=0
            datchr(endian*(cpi+1)+(1-2*endian)*ic)=char(127)
            if(datint.ne.datiu)call fatal('Program error: '//
     &       'Unimplemented machine integer encoding')
         endif
      endif

c     initialize the string and byte positions
      datint=0
      ic=0
      if(addsiz.ne.0)then
         ic=1
         datchr(endian*(cpi+1)+(1-2*endian)*ic)=char(strmax)
      endif

c     pack the characters
      do 1000 str=1,strmax
         ic=ic+1
         if(ic.gt.cpi)then
            dat=dat+1
            if(dat.gt.datdim)goto 9800
            datisu(dat)=datint
            datint=0
            ic=1
         endif
         datchr(endian*(cpi+1)+(1-2*endian)*ic)=strchr(str)
 1000 continue

c     store the pending packed characters
      dat=dat+1
      if(dat.gt.datdim)goto 9800
      datisu(dat)=datint
      goto 9900

c     out of storage
 9800 dat=-1

c     done
 9900 call outsub('pckcs')
      return
      end



      subroutine upckcs(datdim,dat0,datisu,strdim,strmax,strchr)

c     Unpack a character string stored in database datisu into the
c     string strchr.

      implicit none

c input

c     declared number of integers in the database
      integer datdim

c     starting position of the string in datisu
      integer dat0

c     the database as integer storage units, and a couple of selected values
      integer*4 datisu(datdim),datint,datiu
c     characters per integer in datisu as declared above
      integer cpi
      parameter (cpi=4)
c     datint divided into individual characters
      character*1 datchr(cpi)
      equivalence (datchr(1),datint)

c     declared number of characters in strchr
      integer strdim

c input/output

c     number of characters in the string (must be initialized to zero
c     for a stored-length string)
      integer strmax

c output

c     the unpacked string characters
      character*1 strchr(strdim)

c local

c     endianness
      integer endian
      data endian/-1/
      save endian

c     position in datisu
      integer dat

c     position in strchr
      integer str

c     character position in a packed integer
      integer ic

c executable

c     note where we are
      call insub('upckcs')

c     negative number of characters is not possible
      if(strmax.lt.0)
     & call fatal('Program error: negative number of characters')

c     check for endianness
      if(endian.eq.-1)then
         datiu=256
         datiu=127*datiu**(cpi-1)
         ic=1
         endian=0
         datint=0
         datchr(endian*(cpi+1)+(1-2*endian)*ic)=char(127)
         if(datint.ne.datiu)then
            endian=1
            datint=0
            datchr(endian*(cpi+1)+(1-2*endian)*ic)=char(127)
            if(datint.ne.datiu)call fatal('Program error: '//
     &       'Unimplemented machine integer encoding')
         endif
      endif

c     initialize the position in datisu
      dat=dat0
      if(dat.gt.datdim)call fatal('Program error: nonexisting string')

c     initialize the string and byte positions
      datint=datisu(dat)
      ic=0

c     for a stored-length string read the length
      if(strmax.eq.0)then
         ic=1
         strmax=ichar(datchr(endian*(cpi+1)+(1-2*endian)*ic))
         if(strmax.lt.0 .or. strmax.gt.127)
     &    call fatal('Program error: invalid stored-length string')
      endif

c     check size
      if(strmax.gt.strdim)call fatal('Program error: string too long')

c     unpack the characters
      do 1000 str=1,strmax
         ic=ic+1
         if(ic.gt.cpi)then
            dat=dat+1
            if(dat.gt.datdim)
     &       call fatal('Program error: nonexisting end of string')
            datint=datisu(dat)
            ic=1
         endif
         strchr(str)=datchr(endian*(cpi+1)+(1-2*endian)*ic)
 1000 continue

      call outsub('upckcs')
      return
      end



      subroutine srtdat(dupok,datdim,datisu,idim,imax,iundat,
     & iiun,lti,ltmin,ltmax)

c     Sort a database of stored-size strings in datisu and find out
c     where strings of a given length start.

      implicit none

c input

c     whether to allow duplicates
      integer dupok

c     declared number of integers in the database
      integer datdim

c     the database as integer storage units, and a divider for it
      integer*4 datisu(datdim),datdiv
c     characters per integer in datisu as declared above
      integer cpi
      parameter (cpi=4)

c     declared number of strings in the database
      integer idim

c     actual number of strings in the database
      integer imax

c     string position in datisu in terms of the unsorted index iun
      integer iundat(idim)

c input/output

c     in: iun=i;  out: pointer from sorted index i to unsorted index iun
      integer iiun(idim)

c output

c     the string number i at which strings of a given length start
      integer lti(128)

c     the length below which there are no strings
      integer ltmin

c     the length above which there are no strings
      integer ltmax

c local

c     string number
      integer i

c     the string length
      integer lt,lt0

c     position in the database
      integer dat,dat2

c     sort interval
      integer iint
c     whether a change occurred
      integer change

c     number of integers in the packed string
      integer pcklt
c     index in the packed strings
      integer pck

c     unpacked string
      character*1 strchr(127)
      integer strmax,str

c executable

c     note where we are
      call insub('srtdat')

c     find the divider, carefully
      datdiv=256
      datdiv=datdiv**(cpi-1)

c     deal with the case that there are no strings
      if(imax.le.0)then
         ltmin=128
         ltmax=0
         do 100 lt=1,128
            lti(lt)=1
 100     continue
         goto 9900
      endif

c     sort the entries
      iint=imax
 1000 iint=iint/2
      if(iint.eq.0)goto 1900
 1100 change=0
      do 1200 i=1,imax-iint
         dat=iundat(iiun(i))
         dat2=iundat(iiun(i+iint))
         if(datisu(dat2).gt.datisu(dat))goto 1200
         if(datisu(dat2).lt.datisu(dat))goto 1150
         lt=datisu(dat)/datdiv
         pcklt=(lt+1+cpi-1)/cpi
         do 1120 pck=1,pcklt-1
            if(datisu(dat2+pck).gt.datisu(dat+pck))goto 1200
            if(datisu(dat2+pck).lt.datisu(dat+pck))goto 1150
 1120    continue
         if(dupok.eq.0)then
            strmax=0
            call upckcs(datdim,dat,datisu,127,strmax,strchr)
            print1130,(strchr(str),str=1,strmax)
 1130       format('Duplicated: ',127a1)
            call fatal('The above pattern or word occurs twice in '//
     &       'the hyphenation data file')
         endif
         if(iiun(i).le.iiun(i+iint))goto 1200
 1150    change=iiun(i)
         iiun(i)=iiun(i+iint)
         iiun(i+iint)=change
 1200 continue
      if(change.ne.0)goto 1100
      goto 1000
 1900 continue

c     find the lengths
      lt0=0
      ltmin=128
      ltmax=0
      do 3000 i=1,imax
         dat=iundat(iiun(i))
         lt=datisu(dat)/datdiv
         if(lt.le.0)call fatal('Program error: zero length string')
         if(lt.gt.127)
     &    call fatal('Program error: string over 127 characters')
         if(lt.lt.lt0)call fatal(
     &    'Program error: lengths must monotonously increase')
         if(lt.gt.lt0)then
            ltmin=min(ltmin,lt)
            ltmax=max(ltmax,lt)
            do 2500 lt0=lt0+1,lt
               lti(lt0)=i
 2500       continue
            lt0=lt
         endif
 3000 continue
      do 3100 lt0=lt0+1,128
         lti(lt0)=imax+1
 3100 continue

 9900 call outsub('srtdat')
      return
      end



      subroutine fnddat(datdim,datisu,idim,imax,iundat,iiun,
     & lti,ltmin,ltmax,strdim,strmax,strchr,pcklt,i)

c     Find the position of a string in the database of stored-size
c     strings datisu.

      implicit none

c input

c     declared number of integers in the database
      integer datdim

c     the database as integer storage units, and the search string like that
      integer*4 datisu(datdim),pckval(128)
c     characters per integer in datisu as declared above
      integer cpi
      parameter (cpi=4)

c     declared number of strings in the database
      integer idim

c     actual number of strings in the database
      integer imax

c     string position in datisu in terms of the unsorted index iun
      integer iundat(idim)

c     pointer from sorted index i to unsorted index iun
      integer iiun(idim)

c     the string number i at which strings of a given length start
      integer lti(128)

c     the length below which there are no strings
      integer ltmin

c     the length above which there are no strings
      integer ltmax

c     declared number of characters in strchr
      integer strdim

c     number of characters in the string
      integer strmax

c     the string characters
      character*1 strchr(strdim)

c output

c     packed length of the string, if found
      integer pcklt

c     string number of the string or 0 if not found
      integer i

c local

c     packed string index
      integer pck

c     search interval string numbers
      integer ilo,ihi,imid

c     position in the database
      integer dat

c executable

c     note where we are
      call insub('fnddat')

c     assume for now that it is not in the database
      i=0

c     done if no strings so short or long
      if(strmax.lt.ltmin .or. strmax.gt.ltmax)goto 9900

c     set the search interval
      ilo=lti(strmax)
      ihi=lti(strmax+1)-1

c     done if empty
      if(ilo.gt.ihi)goto 9900

c     pack the string
      pcklt=0
      call pckcs(strmax,strchr,1,128,pcklt,pckval)

c     see whether it is out of bounds
      dat=iundat(iiun(ilo))-1
      if(dat+pcklt.gt.datdim)
     & call fatal('Program error: out of the buffer')
      do 100 pck=1,pcklt
         if(pckval(pck).lt.datisu(dat+pck))goto 9900
         if(pckval(pck).gt.datisu(dat+pck))goto 110
 100  continue
      i=ilo
      goto 9900
 110  continue
      dat=iundat(iiun(ihi))-1
      if(dat+pcklt.gt.datdim)
     & call fatal('Program error: out of the buffer, version 2')
      do 200 pck=1,pcklt
         if(pckval(pck).gt.datisu(dat+pck))goto 9900
         if(pckval(pck).lt.datisu(dat+pck))goto 210
 200  continue
      i=ihi
 210  continue

c     perform binary search
 1000 imid=ilo+(ihi-ilo)/2
      if(imid.le.ilo)goto 9900
      dat=iundat(iiun(imid))-1
      if(dat+pcklt.gt.datdim)call fatal(
     & 'Program error: out of the buffer, version 3')
      do 1100 pck=1,pcklt
         if(pckval(pck).lt.datisu(dat+pck))goto 1110
         if(pckval(pck).gt.datisu(dat+pck))goto 1120
 1100 continue
      i=imid
 1110 ihi=imid
      goto 1000
 1120 ilo=imid
      goto 1000

 9900 call outsub('fnddat')
      return
      end



      subroutine numstr(verbos,task,strmax,strchr,num)

c     Map strings to numbers of vice-versa.  This can be used to map
c     strings to strings by changing the mapping halfway.

c     If task is
c        '#': Returns number num corresponding to strchr(1)-strchr(strmax).
c             If there is more than one such number, the lowest one
c             is returned.
c        ':': Returns strchr(1)-strchr(strmax) corresponding to number num.
c             If there is more than one such string, the one that appears
c             first in the data file is returned.
c        other: takes task to be the name of the file with the mapping
c               between numbers and strings and reads in that file.

c     The data file is assumed to consist of data lines and comments.
c     Comments are blank lines or lines that start with a # as first
c     nonblank character.
c     Data lines can be of two allowed forms:
c     1) The data line can start with a nonnegative integer, equal to
c        the value of the number.  This should be followed by one or
c        more whitespace characters and then the string, terminated by
c        whitespace and maybe a trailing comment.
c        Obviously, this does not allow whitespace in the string.  But
c        you might be able to work around that by encodings spaces
c        as DEL characters, depending on conditions.
c        Example data line of this type:
c           163 \pounds # POUND SIGN
c        NOTE: all ASCII characters below 33 are considered whitespace.
c     2) The data line can start with a negative integer, equal to minus
c        the length of the string in bytes.  This must be followed by
c        whitespace, the value of the number, then a *single* whitespace,
c        and then the string.  This allows whitespace and even linebreaks
c        in the string.  Characters behind the string until the end of
c        the line are again ignored.  The end of line must be present.
c        Example data line of this type:
c           -7 163 \pounds # POUND SIGN
c        NOTE: a DOS/Windows new line is *two* whitespaces.
c     The data lines must be ordered so that the numbers are in
c     nondecreasing order.
c     Strings are limited to 127 characters or less.

      implicit none

c input

c     verbosity
      integer verbos

c     the task to perform as described above
      character*(*) task

c input/output:

c     string length (returns zero on task ':' if not found)
      integer strmax

c     the string of characters corresponding to the number
      character*1 strchr(127)

c     the corresponding numeric value (returns -1 on task '#' if not found)
      integer num

c local

c     read mode
      integer rdmode
c     its values
      integer havbeg,inlt,havlt,innum,havnum,instr0,instr,incom
      parameter (havbeg=1,inlt=2,havlt=3,innum=4,havnum=5,instr0=6,
     & instr=7,incom=8)
c     the length of the string to read in
      integer strlt

c     declared number of integers in the database
      integer datdim
      parameter (datdim=50000)
c     actual number of integers in the database
      integer datmax
c     data index
      integer dat
c     the database as integer storage units
      integer*4 datisu(datdim)
c     characters per integer in datisu as declared above
      integer cpi
      parameter (cpi=4)
      save datmax,datisu

c     packed length of the string
      integer pcklt

c     declared number of mappings
      integer mdim
      parameter (mdim=20000)
c     actual number of mappings
      integer mmax
      data mmax/0/
c     index of the sorted mappings
      integer m
c     values of index mun of the unsorted mappings
      integer mun,munlo,munmid,munhi
c     index mun in terms of m
      integer mmun(mdim)
c     the mapped number num in terms of mun
      integer munnum(mdim)
c     mapping location in datisu in terms of mun
      integer mundat(mdim)
c     beginning of mappings of a given length
      integer sltm(128),sltmin,sltmax
      save mmax,mmun,munnum,mundat,sltm,sltmin,sltmax

c     line count in the data file
      integer linct

c     character
      character*1 chr
      integer ichr

c     temporary
      integer itemp

c executable

c     generic formats
 10   format(a,a)
 20   format(a,i6)

c     note where we are
      call insub('numstr')

c     check that we have the needed data
      if(index('#:',task).gt.0 .and. mmax.le.0)
     & call fatal('No number-string mappings read in')

c     dispatch to the right task
      if(task.eq.':')goto 2000
      if(task.eq.'#')goto 3000

c     read the data file
 1000 continue

c     open the input file
      if(verbos.gt.0)print10,'Reading numbers/strings data file ',task
      call openru(task,7)

c     initialize the mappings and storage
      mmax=0
      datmax=0

c     next line
      linct=0
 1100 linct=linct+1
      rdmode=havbeg
      strlt=-1
      strmax=0

c     next character in line
 1110 call getchr(7,chr,*1900,*9110)
      ichr=ichar(chr)

c     handle reading a string of a known length
      if(rdmode.eq.instr0)then
         strmax=strmax+1
         strchr(strmax)=chr
         if(strmax.ge.strlt)then
            ichr=32
            goto 1200
         endif
         goto 1110
      endif

c     handle end of line
      if(ichr.eq.10)then
         if(rdmode.eq.havbeg)goto 1100
         if(rdmode.gt.havbeg .and. rdmode.le.instr0)goto 9120
         if(rdmode.eq.instr)goto 1200
         if(rdmode.eq.incom)goto 1100
         call fatal('Program error: no such read mode')
      endif

c     ignore all else for comments
      if(rdmode.eq.incom)goto 1110

c     handle other read modes
      if(rdmode.eq.havbeg)then
c        process the new character at the start of the line
         if(ichr.le.32)goto 1110
         if(ichr.eq.ichar('#'))then
            rdmode=incom
            goto 1110
         endif
         if(ichr.eq.ichar('-'))then
            rdmode=inlt
            strlt=0
            goto 1110
         endif
         if(ichr.ge.48 .and. ichr.le.57)then
            rdmode=innum
            num=ichr-48
            goto 1110
         endif
         goto 9150
      else if(rdmode.eq.inlt)then 
c        process the new character while reading the string length
         if(ichr.ge.48 .and. ichr.le.57)then
            strlt=strlt+ichr-48
            if(strlt.gt.127)goto 9140
            goto 1110
         endif
         if(ichr.le.32)then
            if(strlt.le.0)goto 9160
            rdmode=havlt
            goto 1110
         endif
         goto 9150
      else if(rdmode.eq.havlt)then 
c        process the new character while having the string length
         if(ichr.le.32)goto 1110
         if(ichr.ge.48 .and. ichr.le.57)then
            num=ichr-48
            rdmode=innum
            goto 1110
         endif
         goto 9150
      else if(rdmode.eq.innum)then
c        process the new character while reading the number
         if(ichr.ge.48 .and. ichr.le.57)then
            if(num.ge.214748364)goto 9170
            num=num*10+ichr-48
            goto 1110
         endif
         if(ichr.le.32)then
c           next mapping
            if(mmax.ge.mdim)goto 9210
            mmax=mmax+1
            if(mmax.gt.1)then
               if(munnum(mmax-1).gt.num)goto 9180
            endif
            munnum(mmax)=num
            mmun(mmax)=mmax
            mundat(mmax)=datmax+1
            rdmode=havnum
            if(strlt.ge.0)rdmode=instr0
            goto 1110
         endif
         goto 9150
      else if(rdmode.eq.havnum)then
c        process the new character while having the number
         if(ichr.le.32)goto 1110
         rdmode=instr
         strmax=1
         strchr(strmax)=chr
         goto 1110
      else if(rdmode.eq.instr)then 
c        process the new character while reading the string
         if(ichr.le.32)goto 1200
         strmax=strmax+1
         if(strmax.gt.127)goto 9140
         strchr(strmax)=chr
         goto 1110
      else
c        that is it
         call fatal('Program error: no other read mode')
      endif

c     add the number-string mapping to the data
 1200 continue

c     store the characters
      dat=datmax
      call pckcs(strmax,strchr,1,datdim,dat,datisu)
      if(dat.lt.0)goto 9220
      datmax=dat

c     go read the next character in the data file
      if(ichr.eq.-1)goto 1910
      if(ichr.eq.10)goto 1100
      rdmode=incom
      goto 1110

c     close the output file
 1900 ichr=-1
      call closu(7,task)
      if(rdmode.gt.havbeg .and. rdmode.le.instr0)goto 9130
      if(rdmode.eq.instr)goto 1200
 1910 continue

c     sort the mappings
      call srtdat(1,datdim,datisu,mdim,mmax,mundat,
     & mmun,sltm,sltmin,sltmax)

c     show statistics
      if(verbos.gt.0)then
         print20,'number of mappings in the file:',mmax
         print20,'smallest number:',munnum(1)
         print20,'largest number:',munnum(mmax)
         print20,'length of shortest string:',sltmin
         print20,'length of longest string:',sltmax
      endif

c     task is done
      goto 9000

c     return the string corresponding to a number
 2000 continue

c     initialize to not found
      mun=-1

c     see whether num is out of bounds
      munlo=1
      if(munnum(munlo).gt.num)goto 2900
      if(munnum(munlo).eq.num)then
         mun=munlo
         goto 2900
      endif
      munhi=mmax
      if(munnum(munhi).lt.num)goto 2900
      if(munnum(munhi).eq.num)mun=munhi

c     perform binary search
 2100 munmid=munlo+(munhi-munlo)/2
      if(munmid.le.munlo)goto 2900
      itemp=munnum(munmid)
      if(itemp.lt.num)then
         munlo=munmid
      else
         munhi=munmid
         if(itemp.eq.num)mun=munhi
      endif
      goto 2100

c     return string mun or strmax=0
 2900 strmax=0
      if(mun.lt.0)goto 9000
      call upckcs(datdim,mundat(mun),datisu,127,strmax,strchr)

c     task is done
      goto 9000

c     return the number corresponding to a string
 3000 continue

c     initialize to not found
      num=-1
      if(strmax.le.0 .or. strmax.gt.127)goto 9000

c     try to find the string
      call fnddat(datdim,datisu,mdim,mmax,mundat,mmun,
     & sltm,sltmin,sltmax,127,strmax,strchr,pcklt,m)

c     return -1 if there is no such string
      if(m.le.0)goto 9000

c     return the number
      mun=mmun(m)
      num=munnum(mun)

c     task is done
      goto 9000

c     return
 9000 call outsub('numstr')
      return

c     various errors with the data file
 9100 format('*** Problem reading mappings file ',a,' line',i6)
 9110 print9100,task,linct
      call fatal('Disk error reading the file')
 9120 print9100,task,linct
      call fatal('Premature end of line')
 9130 print9100,task,linct
      call fatal('Premature end of file')
 9140 print9100,task,linct
      call fatal('The string cannot exceed 127 bytes')
 9150 print9100,task,linct
      call fatal('Invalid line format')
 9160 print9100,task,linct
      call fatal('The string length cannot be zero')
 9170 print9100,task,linct
      call fatal('The number is too big')
 9180 print9100,task,linct
      call fatal('The numbers may not decrease in value')
c     various errors processing the mappings
 9200 format('Current mapping',i6,' (limit',i6,'), string storage',i6,
     & ' (limit',i6,')')
 9210 print9200,mmax,mdim,datmax,datdim
      call fatal('Too many number-string mappings')
 9220 print9200,mmax,mdim,datmax,datdim
      call fatal('Available string storage exceeded')

      end
